@@ -16,18 +16,22 @@
 import argparse
 import datetime
 import dateutil
+import fcntl
+import functools
 import logging
 import os
-import Queue
 import re
+import socket
 import subprocess
 import sys
 import textwrap
 import threading
-import urlparse
 import warnings
 import webbrowser
 
+import six
+from six.moves import queue
+from six.moves.urllib import parse as urlparse
 import sqlalchemy.exc
 import urwid
 
@@ -133,6 +137,43 @@ class StatusHeader(urwid.WidgetWrap):
             self.sync_widget.set_text(u' Sync: %i' % self._sync)
 
 
+class BreadCrumbBar(urwid.WidgetWrap):
+    BREADCRUMB_SYMBOL = u'\N{BLACK RIGHT-POINTING SMALL TRIANGLE}'
+    BREADCRUMB_WIDTH = 25
+
+    def __init__(self):
+        self.prefix_text = urwid.Text(u' \N{WATCH}  ')
+        self.breadcrumbs = urwid.Columns([], dividechars=3)
+        self.display_widget = urwid.Columns(
+            [('pack', self.prefix_text), self.breadcrumbs])
+        super(BreadCrumbBar, self).__init__(self.display_widget)
+
+    def _get_breadcrumb_text(self, screen):
+        title = getattr(screen, 'short_title', None)
+        if not title:
+            title = getattr(screen, 'title', str(screen))
+        text = "%s %s" % (BreadCrumbBar.BREADCRUMB_SYMBOL, title)
+        if len(text) > 23:
+            text = "%s..." % text[:20]
+        return urwid.Text(text, wrap='clip')
+
+    def _get_breadcrumb_column_options(self):
+        return self.breadcrumbs.options("given", BreadCrumbBar.BREADCRUMB_WIDTH)
+
+    def _update(self, screens):
+        breadcrumb_contents = []
+        for screen in screens:
+            breadcrumb_contents.append((
+                self._get_breadcrumb_text(screen),
+                self._get_breadcrumb_column_options()))
+        self.breadcrumbs.contents = breadcrumb_contents
+        # Update focus so we always have the right end of the breadcrumb trail
+        # in view. Urwid will gracefully handle clipping from the left when
+        # there is overflow.as trail grows, shrinks, or screen is resized.
+        if len(self.breadcrumbs.contents):
+            self.breadcrumbs.focus_position = len(self.breadcrumbs.contents) - 1
+
+
 class SearchDialog(mywid.ButtonDialog):
     signals = ['search', 'cancel']
     def __init__(self, app, default):
@@ -185,12 +226,30 @@ class BackgroundBrowser(webbrowser.GenericBrowser):
         except OSError:
             return False
 
+class ProjectCache(object):
+    def __init__(self):
+        self.projects = {}
+
+    def get(self, project):
+        if project.key not in self.projects:
+            self.projects[project.key] = dict(
+                unreviewed_changes = len(project.unreviewed_changes),
+                open_changes = len(project.open_changes),
+            )
+        return self.projects[project.key]
+
+    def clear(self, project):
+        if project.key in self.projects:
+            del self.projects[project.key]
+
 class App(object):
     simple_change_search = re.compile('^(\d+|I[a-fA-F0-9]{40})$')
 
-    def __init__(self, server=None, palette='default', keymap='default',
-                 debug=False, verbose=False, disable_sync=False,
-                 fetch_missing_refs=False, path=config.DEFAULT_CONFIG_PATH):
+    def __init__(self, server=None, palette='default',
+                 keymap='default', debug=False, verbose=False,
+                 disable_sync=False, disable_background_sync=False,
+                 fetch_missing_refs=False,
+                 path=config.DEFAULT_CONFIG_PATH):
         self.server = server
         self.config = config.Config(server, palette, keymap, path)
         if debug:
@@ -216,6 +275,14 @@ class App(object):
         self.log = logging.getLogger('gertty.App')
         self.log.debug("Starting")
 
+        self.lock_fd = open(self.config.lock_file, 'w')
+        try:
+            fcntl.lockf(self.lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
+        except IOError:
+            print("error: another instance of gertty is running for: %s" % self.config.server['name'])
+            sys.exit(1)
+
+        self.project_cache = ProjectCache()
         self.ring = mywid.KillRing()
         self.input_buffer = []
         webbrowser.register('xdg-open', None, BackgroundBrowser("xdg-open"))
@@ -224,21 +291,33 @@ class App(object):
         self.config.keymap.updateCommandMap()
         self.search = search.SearchCompiler(self.config.username)
         self.db = db.Database(self, self.config.dburi, self.search)
-        self.sync = sync.Sync(self)
+        self.sync = sync.Sync(self, disable_background_sync)
 
-        self.screens = []
         self.status = StatusHeader(self)
         self.header = urwid.AttrMap(self.status, 'header')
+        self.screens = urwid.MonitoredList()
+        self.breadcrumbs = BreadCrumbBar()
+        self.screens.set_modified_callback(
+            functools.partial(self.breadcrumbs._update, self.screens))
+        if self.config.breadcrumbs:
+            self.footer = urwid.AttrMap(self.breadcrumbs, 'footer')
+        else:
+            self.footer = None
         screen = view_project_list.ProjectListView(self)
         self.status.update(title=screen.title)
         self.updateStatusQueries()
-        self.loop = urwid.MainLoop(screen, palette=self.config.palette.getPalette(),
-                                   unhandled_input=self.unhandledInput)
+        self.frame = urwid.Frame(body=screen, footer=self.footer)
+        self.loop = urwid.MainLoop(self.frame, palette=self.config.palette.getPalette(),
+                                   handle_mouse=self.config.handle_mouse,
+                                   unhandled_input=self.unhandledInput,
+                                   input_filter=self.inputFilter)
 
         self.sync_pipe = self.loop.watch_pipe(self.refresh)
-        self.error_queue = Queue.Queue()
+        self.error_queue = queue.Queue()
         self.error_pipe = self.loop.watch_pipe(self._errorPipeInput)
         self.logged_warnings = set()
+        self.command_pipe = self.loop.watch_pipe(self._commandPipeInput)
+        self.command_queue = queue.Queue()
 
         warnings.showwarning = self._showWarning
 
@@ -251,6 +330,9 @@ class App(object):
 
         self.loop.screen.tty_signal_keys(start='undefined', stop='undefined')
         #self.loop.screen.set_terminal_properties(colors=88)
+
+        self.startSocketListener()
+
         if not disable_sync:
             self.sync_thread = threading.Thread(target=self.sync.run, args=(self.sync_pipe,))
             self.sync_thread.daemon = True
@@ -277,6 +359,35 @@ class App(object):
 
         self.popup(dialog)
 
+    def startSocketListener(self):
+        if os.path.exists(self.config.socket_path):
+            os.unlink(self.config.socket_path)
+        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+        self.socket.bind(self.config.socket_path)
+        self.socket.listen(1)
+        self.socket_thread = threading.Thread(target=self._socketListener)
+        self.socket_thread.daemon = True
+        self.socket_thread.start()
+
+    def _socketListener(self):
+        while True:
+            try:
+                s, addr = self.socket.accept()
+                self.log.debug("Accepted socket connection %s" % (s,))
+                buf = ''
+                while True:
+                    buf += s.recv(1)
+                    if buf[-1] == '\n':
+                        break
+                buf = buf.strip()
+                self.log.debug("Received %s from socket" % (buf,))
+                s.close()
+                parts = buf.split()
+                self.command_queue.put((parts[0], parts[1:]))
+                os.write(self.command_pipe, six.b('command\n'))
+            except Exception:
+                self.log.exception("Exception in socket handler")
+
     def clearInputBuffer(self):
         if self.input_buffer:
             self.input_buffer = []
@@ -286,9 +397,9 @@ class App(object):
         self.log.debug("Changing screen to %s" % (widget,))
         self.status.update(error=False, title=widget.title)
         if push:
-            self.screens.append(self.loop.widget)
+            self.screens.append(self.frame.body)
         self.clearInputBuffer()
-        self.loop.widget = widget
+        self.frame.body = widget
 
     def backScreen(self, target_widget=None):
         if not self.screens:
@@ -301,7 +412,7 @@ class App(object):
         if hasattr(widget, 'title'):
             self.status.update(title=widget.title)
         self.clearInputBuffer()
-        self.loop.widget = widget
+        self.frame.body = widget
         self.refresh(force=True)
 
     def findChangeList(self):
@@ -315,10 +426,10 @@ class App(object):
         while self.screens:
             widget = self.screens.pop()
             self.clearInputBuffer()
-            self.loop.widget = widget
+            self.frame.body = widget
 
     def refresh(self, data=None, force=False):
-        widget = self.loop.widget
+        widget = self.frame.body
         while isinstance(widget, urwid.Overlay):
             widget = widget.contents[0][0]
         interested = force
@@ -330,7 +441,7 @@ class App(object):
                     interested = True
                 if hasattr(event, 'held_changed') and event.held_changed:
                     invalidate = True
-        except Queue.Empty:
+        except queue.Empty:
             pass
         if interested:
             widget.refresh()
@@ -345,28 +456,41 @@ class App(object):
 
     def popup(self, widget,
               relative_width=50, relative_height=25,
-              min_width=20, min_height=8):
+              min_width=20, min_height=8,
+              width=None, height=None):
         self.clearInputBuffer()
-        overlay = urwid.Overlay(widget, self.loop.widget,
-                                'center', ('relative', relative_width),
-                                'middle', ('relative', relative_height),
+        if width is None:
+            width = ('relative', relative_width)
+        if height is None:
+            height = ('relative', relative_height)
+        overlay = urwid.Overlay(widget, self.frame.body,
+                                'center', width,
+                                'middle', height,
                                 min_width=min_width, min_height=min_height)
-        self.log.debug("Overlaying %s on screen %s" % (widget, self.loop.widget))
-        self.screens.append(self.loop.widget)
-        self.loop.widget = overlay
+        if hasattr(widget, 'title'):
+            overlay.title = widget.title
+        self.log.debug("Overlaying %s on screen %s" % (widget, self.frame.body))
+        self.screens.append(self.frame.body)
+        self.frame.body = overlay
+
+    def getGlobalCommands(self):
+        return list(mywid.GLOBAL_HELP)
+
+    def getGlobalHelp(self):
+        keys =  [(k, self.config.keymap.formatKeys(k), t) for (k, t) in self.getGlobalCommands()]
+        for d in self.config.dashboards.values():
+            keys.append(('', d['key'], d['name']))
+        return keys
 
     def help(self):
-        if not hasattr(self.loop.widget, 'help'):
+        if not hasattr(self.frame.body, 'help'):
             return
-        global_help = [(self.config.keymap.formatKeys(k), t)
-                       for (k, t) in mywid.GLOBAL_HELP]
-        for d in self.config.dashboards.values():
-            global_help.append((keymap.formatKey(d['key']), d['name']))
+        global_help = self.getGlobalHelp()
         parts = [('Global Keys', global_help),
-                 ('This Screen', self.loop.widget.help())]
+                 ('This Screen', self.frame.body.help())]
         keylen = 0
         for title, items in parts:
-            for keys, text in items:
+            for cmd, keys, text in items:
                 keylen = max(len(keys), keylen)
         text = ''
         for title, items in parts:
@@ -374,7 +498,7 @@ class App(object):
                 text += '\n'
             text += title+'\n'
             text += '%s\n' % ('='*len(title),)
-            for keys, cmdtext in items:
+            for cmd, keys, cmdtext in items:
                 text += '{keys:{width}} {text}\n'.format(
                     keys=keys, width=keylen, text=cmdtext)
         dialog = mywid.MessageDialog('Help for %s' % version(), text)
@@ -392,7 +516,7 @@ class App(object):
         self.popup(dialog, min_width=76, min_height=len(lines)+4)
 
     def _syncOneChangeFromQuery(self, query):
-        number = changeid = None
+        number = changeid = restid = None
         if query.startswith("change:"):
             number = query.split(':')[1].strip()
             try:
@@ -404,29 +528,41 @@ class App(object):
             return
         with self.db.getSession() as session:
             if number:
-                change = session.getChangeByNumber(number)
+                changes = [session.getChangeByNumber(number)]
             elif changeid:
-                change = session.getChangeByChangeID(changeid)
-            change_key = change and change.key or None
-        if change_key is None:
+                changes = session.getChangesByChangeID(changeid)
+            change_keys = [c.key for c in changes if c]
+            restids = [c.id for c in changes if c]
+        if not change_keys:
             if self.sync.offline:
                 raise Exception('Can not sync change while offline.')
-            task = sync.SyncChangeByNumberTask(number or changeid, sync.HIGH_PRIORITY)
-            self.sync.submitTask(task)
-            succeeded = task.wait(300)
-            if not succeeded:
-                raise Exception('Unable to find change.')
-            for subtask in task.tasks:
-                succeeded = subtask.wait(300)
+            dialog = mywid.SystemMessage("Syncing change...")
+            self.popup(dialog, width=40, height=6)
+            self.loop.draw_screen()
+            try:
+                task = sync.SyncChangeByNumberTask(number or changeid, sync.HIGH_PRIORITY)
+                self.sync.submitTask(task)
+                succeeded = task.wait(300)
                 if not succeeded:
-                    raise Exception('Unable to sync change.')
+                    raise Exception('Unable to find change.')
+                for subtask in task.tasks:
+                    succeeded = subtask.wait(300)
+                    if not succeeded:
+                        raise Exception('Unable to sync change.')
+            finally:
+                # Remove "syncing..." popup
+                self.backScreen()
             with self.db.getSession() as session:
                 if number:
-                    change = session.getChangeByNumber(number)
+                    changes = [session.getChangeByNumber(number)]
                 elif changeid:
-                    change = session.getChangeByChangeID(changeid)
-                change_key = change and change.key or None
-        if change_key is None:
+                    changes = session.getChangesByChangeID(changeid)
+                change_keys = [c.key for c in changes if c]
+        elif restids:
+            for restid in restids:
+                task = sync.SyncChangeTask(restid, sync.HIGH_PRIORITY)
+                self.sync.submitTask(task)
+        if not change_keys:
             raise Exception('Change is not in local database.')
 
     def doSearch(self, query):
@@ -521,6 +657,13 @@ class App(object):
         self.popup(dialog, min_height=min_height)
         return None
 
+    def inputFilter(self, keys, raw):
+        if 'window resize' in keys:
+            m = getattr(self.frame.body, 'onResize', None)
+            if m:
+                m()
+        return keys
+
     def unhandledInput(self, key):
         # get commands from buffer
         keys = self.input_buffer + [key]
@@ -540,11 +683,25 @@ class App(object):
             self.doSearch("is:held")
         elif key in self.config.dashboards:
             d = self.config.dashboards[key]
-            view = view_change_list.ChangeListView(self, d['query'], d['name'])
+            view = view_change_list.ChangeListView(self, d['query'], d['name'],
+                                                   sort_by=d.get('sort-by'),
+                                                   reverse=d.get('reverse'))
             self.changeScreen(view)
         elif keymap.FURTHER_INPUT in commands:
             self.input_buffer.append(key)
-            self.status.update(message=''.join(self.input_buffer))
+            msg = ''.join(self.input_buffer)
+            commands = dict(self.getGlobalCommands())
+            if hasattr(self.frame.body, 'getCommands'):
+                commands.update(dict(self.frame.body.getCommands()))
+            further_commands = self.config.keymap.getFurtherCommands(keys)
+            completions = []
+            for (key, cmds) in further_commands:
+                for cmd in cmds:
+                    if cmd in commands:
+                        completions.append(key)
+            completions = ' '.join(completions)
+            msg = '%s: %s' % (msg, completions)
+            self.status.update(message=msg)
             return
         self.clearInputBuffer()
 
@@ -576,12 +733,25 @@ class App(object):
         # nearly un-actionable.
         if category == requestsexceptions.InsecurePlatformWarning:
             return
+        if category == requestsexceptions.SNIMissingWarning:
+            return
         # Disable InsecureRequestWarning when certificate validation is disabled
         if not self.config.verify_ssl:
             if category == requestsexceptions.InsecureRequestWarning:
                 return
         self.error_queue.put(('Warning', m))
-        os.write(self.error_pipe, 'error\n')
+        os.write(self.error_pipe, six.b('error\n'))
+
+    def _commandPipeInput(self, data=None):
+        (command, data) = self.command_queue.get()
+        if command == 'open':
+            url = data[0]
+            self.log.debug("Opening URL %s" % (url,))
+            result = self.parseInternalURL(url)
+            if result is not None:
+                self.openInternalURL(result)
+        else:
+            self.log.error("Unable to parse command %s with data %s" % (command, data))
 
     def toggleHeldChange(self, change_key):
         with self.db.getSession() as session:
@@ -668,6 +838,7 @@ class App(object):
             message_key = draft_message.key
         if upload:
             change.reviewed = True
+            self.project_cache.clear(change.project)
         if submit:
             change.status = 'SUBMITTED'
             change.pending_status = True
@@ -682,13 +853,28 @@ def version():
 class PrintKeymapAction(argparse.Action):
     def __call__(self, parser, namespace, values, option_string=None):
         for cmd in sorted(keymap.DEFAULT_KEYMAP.keys()):
-            print cmd.replace(' ', '-')
+            print(cmd.replace(' ', '-'))
         sys.exit(0)
 
 class PrintPaletteAction(argparse.Action):
     def __call__(self, parser, namespace, values, option_string=None):
         for attr in sorted(palette.DEFAULT_PALETTE.keys()):
-            print attr
+            print(attr)
+        sys.exit(0)
+
+class OpenChangeAction(argparse.Action):
+    def __call__(self, parser, namespace, values, option_string=None):
+        cf = config.Config(namespace.server, namespace.palette,
+                           namespace.keymap, namespace.path)
+        url = values[0]
+        result = urlparse.urlparse(values[0])
+        if not url.startswith(cf.url):
+            print('Supplied URL must start with %s' % (cf.url,))
+            sys.exit(1)
+
+        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+        s.connect(cf.socket_path)
+        s.sendall('open %s\n' % url)
         sys.exit(0)
 
 def main():
@@ -703,6 +889,8 @@ def main():
                         help='enable debug logging')
     parser.add_argument('--no-sync', dest='no_sync', action='store_true',
                         help='disable remote syncing')
+    parser.add_argument('--debug-sync', dest='debug_sync', action='store_true',
+                        help='disable most background sync tasks for debugging')
     parser.add_argument('--fetch-missing-refs', dest='fetch_missing_refs',
                         action='store_true',
                         help='fetch any refs missing from local repos')
@@ -710,6 +898,9 @@ def main():
                         help='print the keymap command names to stdout')
     parser.add_argument('--print-palette', nargs=0, action=PrintPaletteAction,
                         help='print the palette attribute names to stdout')
+    parser.add_argument('--open', nargs=1, action=OpenChangeAction,
+                        metavar='URL',
+                        help='open the given URL in a running Gertty')
     parser.add_argument('--version', dest='version', action='version',
                         version=version(),
                         help='show Gertty\'s version')
@@ -721,7 +912,7 @@ def main():
                         help='the server to use (as specified in config file)')
     args = parser.parse_args()
     g = App(args.server, args.palette, args.keymap, args.debug, args.verbose,
-            args.no_sync, args.fetch_missing_refs, args.path)
+            args.no_sync, args.debug_sync, args.fetch_missing_refs, args.path)
     g.run()
 
 
