@@ -21,9 +21,9 @@ try:
     import ordereddict
 except:
     pass
-import urlparse
 import yaml
 
+from six.moves.urllib import parse as urlparse
 import voluptuous as v
 
 import gertty.commentlink
@@ -46,12 +46,16 @@ class ConfigSchema(object):
               'ssl-ca-path': str,
               'dburi': str,
               v.Required('git-root'): str,
+              'git-url': str,
               'log-file': str,
-              'auth-type': str,
+              'socket': str,
+              'auth-type': v.Any('basic', 'digest', 'form'),
               }
 
     servers = [server]
 
+    sort_by = v.Any('number', 'updated', 'last-seen')
+
     text_replacement = {'text': v.Any(str,
                                       {'color': str,
                                        v.Required('text'): str})}
@@ -77,6 +81,8 @@ class ConfigSchema(object):
 
     dashboard = {v.Required('name'): str,
                  v.Required('query'): str,
+                 v.Optional('sort-by'): sort_by,
+                 v.Optional('reverse'): bool,
                  v.Required('key'): str}
 
     dashboards = [dashboard]
@@ -94,7 +100,7 @@ class ConfigSchema(object):
 
     hide_comments = [hide_comment]
 
-    change_list_options = {'sort-by': v.Any('number', 'updated'),
+    change_list_options = {'sort-by': sort_by,
                            'reverse': bool}
 
     keymap = {v.Required('name'): str,
@@ -116,6 +122,8 @@ class ConfigSchema(object):
                            'hide-comments': self.hide_comments,
                            'thread-changes': bool,
                            'display-times-in-utc': bool,
+                           'handle-mouse': bool,
+                           'breadcrumbs': bool,
                            'change-list-options': self.change_list_options,
                            'expire-age': str,
                            })
@@ -157,9 +165,6 @@ class Config(object):
                     "Permissions are: {}".format(self.path, oct(mode)))
                 exit(1)
         self.auth_type = server.get('auth-type', 'digest')
-        auth_types = ['digest', 'basic']
-        if self.auth_type not in auth_types:
-            self.auth_type = 'digest'
         self.verify_ssl = server.get('verify-ssl', True)
         if not self.verify_ssl:
             os.environ['GIT_SSL_NO_VERIFY']='true'
@@ -171,10 +176,18 @@ class Config(object):
             # And this is to allow Git callouts
             os.environ['GIT_SSL_CAINFO'] = self.ssl_ca_path
         self.git_root = os.path.expanduser(server['git-root'])
+        git_url = server.get('git-url', self.url + 'p/')
+        if not git_url.endswith('/'):
+            git_url += '/'
+        self.git_url = git_url
         self.dburi = server.get('dburi',
                                 'sqlite:///' + os.path.expanduser('~/.gertty.db'))
+        socket_path = server.get('socket', '~/.gertty.sock')
+        self.socket_path = os.path.expanduser(socket_path)
         log_file = server.get('log-file', '~/.gertty.log')
         self.log_file = os.path.expanduser(log_file)
+        lock_file = server.get('lock-file', '~/.gertty.%s.lock' % server['name'])
+        self.lock_file = os.path.expanduser(lock_file)
 
         self.palettes = {'default': gertty.palette.Palette({}),
                          'light': gertty.palette.Palette(gertty.palette.LIGHT_PALETTE),
@@ -186,7 +199,8 @@ class Config(object):
                 self.palettes[p['name']].update(p)
         self.palette = self.palettes[self.config.get('palette', palette)]
 
-        self.keymaps = {'default': gertty.keymap.KeyMap({})}
+        self.keymaps = {'default': gertty.keymap.KeyMap({}),
+                        'vi': gertty.keymap.KeyMap(gertty.keymap.VI_KEYMAP)}
         for p in self.config.get('keymaps', []):
             if p['name'] not in self.keymaps:
                 self.keymaps[p['name']] = gertty.keymap.KeyMap(p)
@@ -211,6 +225,7 @@ class Config(object):
         self.dashboards = OrderedDict()
         for d in self.config.get('dashboards', []):
             self.dashboards[d['key']] = d
+            self.dashboards[d['key']]
 
         self.reviewkeys = OrderedDict()
         for k in self.config.get('reviewkeys', []):
@@ -222,6 +237,8 @@ class Config(object):
 
         self.thread_changes = self.config.get('thread-changes', True)
         self.utc = self.config.get('display-times-in-utc', False)
+        self.breadcrumbs = self.config.get('breadcrumbs', True)
+        self.handle_mouse = self.config.get('handle-mouse', True)
 
         change_list_options = self.config.get('change-list-options', {})
         self.change_list_options = {
@@ -238,11 +255,11 @@ class Config(object):
 
     def printSample(self):
         filename = 'share/gertty/examples'
-        print """Gertty requires a configuration file at ~/.gertty.yaml
+        print("""Gertty requires a configuration file at ~/.gertty.yaml
 If the file contains a password then permissions must be set to 0600.
 
 Several sample configuration files were installed with Gertty and are
 available in %s in the root of the installation.
 
 For more information, please see the README.
-""" % (filename,)
+""" % (filename,))
