@@ -20,10 +20,11 @@ import threading
 
 import alembic
 import alembic.config
+import six
 import sqlalchemy
-from sqlalchemy import create_engine, MetaData, Table, Column, Integer, String, Boolean, DateTime, Text
+from sqlalchemy import create_engine, MetaData, Table, Column, Integer, String, Boolean, DateTime, Text, UniqueConstraint
 from sqlalchemy.schema import ForeignKey
-from sqlalchemy.orm import mapper, sessionmaker, relationship, scoped_session
+from sqlalchemy.orm import mapper, sessionmaker, relationship, scoped_session, joinedload
 from sqlalchemy.orm.session import Session
 from sqlalchemy.sql import exists
 from sqlalchemy.sql.expression import and_
@@ -43,6 +44,20 @@ branch_table = Table(
     Column('project_key', Integer, ForeignKey("project.key"), index=True),
     Column('name', String(255), index=True, nullable=False),
     )
+topic_table = Table(
+    'topic', metadata,
+    Column('key', Integer, primary_key=True),
+    Column('name', String(255), index=True, nullable=False),
+    Column('sequence', Integer, index=True, unique=True, nullable=False),
+    )
+project_topic_table = Table(
+    'project_topic', metadata,
+    Column('key', Integer, primary_key=True),
+    Column('project_key', Integer, ForeignKey("project.key"), index=True),
+    Column('topic_key', Integer, ForeignKey("topic.key"), index=True),
+    Column('sequence', Integer, nullable=False),
+    UniqueConstraint('topic_key', 'sequence', name='topic_key_sequence_const'),
+    )
 change_table = Table(
     'change', metadata,
     Column('key', Integer, primary_key=True),
@@ -66,6 +81,13 @@ change_table = Table(
     Column('pending_starred', Boolean, index=True, nullable=False),
     Column('pending_status', Boolean, index=True, nullable=False),
     Column('pending_status_message', Text),
+    Column('last_seen', DateTime, index=True),
+    )
+change_conflict_table = Table(
+    'change_conflict', metadata,
+    Column('key', Integer, primary_key=True),
+    Column('change1_key', Integer, ForeignKey("change.key"), index=True),
+    Column('change2_key', Integer, ForeignKey("change.key"), index=True),
     )
 revision_table = Table(
     'revision', metadata,
@@ -200,6 +222,35 @@ class Branch(object):
         self.project_key = project.key
         self.name = name
 
+class ProjectTopic(object):
+    def __init__(self, project, topic, sequence):
+        self.project_key = project.key
+        self.topic_key = topic.key
+        self.sequence = sequence
+
+class Topic(object):
+    def __init__(self, name, sequence):
+        self.name = name
+        self.sequence = sequence
+
+    def addProject(self, project):
+        session = Session.object_session(self)
+        seq = max([x.sequence for x in self.project_topics] + [0])
+        pt = ProjectTopic(project, self, seq+1)
+        self.project_topics.append(pt)
+        self.projects.append(project)
+        session.add(pt)
+        session.flush()
+
+    def removeProject(self, project):
+        session = Session.object_session(self)
+        for pt in self.project_topics:
+            if pt.project_key == project.key:
+                self.project_topics.remove(pt)
+                session.delete(pt)
+        self.projects.remove(project)
+        session.flush()
+
 class Change(object):
     def __init__(self, project, id, owner, number, branch, change_id,
                  subject, created, updated, status, topic=None,
@@ -322,6 +373,30 @@ class Change(object):
                 owner_name = self.owner.email
         return owner_name
 
+    @property
+    def conflicts(self):
+        return tuple(set(self.conflicts1 + self.conflicts2))
+
+    def addConflict(self, other):
+        session = Session.object_session(self)
+        if other in self.conflicts1 or other in self.conflicts2:
+            return
+        if self in other.conflicts1 or self in other.conflicts2:
+            return
+        self.conflicts1.append(other)
+        session.flush()
+        session.expire(other, attribute_names=['conflicts2'])
+
+    def delConflict(self, other):
+        session = Session.object_session(self)
+        if other in self.conflicts1:
+            self.conflicts1.remove(other)
+            session.flush()
+            session.expire(other, attribute_names=['conflicts2'])
+        if self in other.conflicts1:
+            other.conflicts1.remove(self)
+            session.flush()
+            session.expire(self, attribute_names=['conflicts2'])
 
 class Revision(object):
     def __init__(self, change, number, message, commit, parent,
@@ -506,30 +581,52 @@ class File(object):
 
 mapper(Account, account_table)
 mapper(Project, project_table, properties=dict(
-        branches=relationship(Branch, backref='project',
-                              order_by=branch_table.c.name,
-                              cascade='all, delete-orphan'),
-        changes=relationship(Change, backref='project',
-                             order_by=change_table.c.number,
-                             cascade='all, delete-orphan'),
-        unreviewed_changes=relationship(Change,
-                                        primaryjoin=and_(project_table.c.key==change_table.c.project_key,
-                                                         change_table.c.hidden==False,
-                                                         change_table.c.status!='MERGED',
-                                                         change_table.c.status!='ABANDONED',
-                                                         change_table.c.reviewed==False),
-                                        order_by=change_table.c.number,
-                                        ),
-        open_changes=relationship(Change,
-                                  primaryjoin=and_(project_table.c.key==change_table.c.project_key,
-                                                   change_table.c.status!='MERGED',
-                                                   change_table.c.status!='ABANDONED'),
-                                  order_by=change_table.c.number,
-                                  ),
-        ))
+    branches=relationship(Branch, backref='project',
+                          order_by=branch_table.c.name,
+                          cascade='all, delete-orphan'),
+    changes=relationship(Change, backref='project',
+                         order_by=change_table.c.number,
+                         cascade='all, delete-orphan'),
+    topics=relationship(Topic,
+                        secondary=project_topic_table,
+                        order_by=topic_table.c.name,
+                        viewonly=True),
+    unreviewed_changes=relationship(Change,
+                                    primaryjoin=and_(project_table.c.key==change_table.c.project_key,
+                                                     change_table.c.hidden==False,
+                                                     change_table.c.status!='MERGED',
+                                                     change_table.c.status!='ABANDONED',
+                                                     change_table.c.reviewed==False),
+                                    order_by=change_table.c.number,
+                                ),
+    open_changes=relationship(Change,
+                              primaryjoin=and_(project_table.c.key==change_table.c.project_key,
+                                               change_table.c.status!='MERGED',
+                                               change_table.c.status!='ABANDONED'),
+                              order_by=change_table.c.number,
+                          ),
+))
 mapper(Branch, branch_table)
+mapper(Topic, topic_table, properties=dict(
+    projects=relationship(Project,
+                          secondary=project_topic_table,
+                          order_by=project_table.c.name,
+                          viewonly=True),
+    project_topics=relationship(ProjectTopic),
+))
+mapper(ProjectTopic, project_topic_table)
 mapper(Change, change_table, properties=dict(
         owner=relationship(Account),
+        conflicts1=relationship(Change,
+                                secondary=change_conflict_table,
+                                primaryjoin=change_table.c.key==change_conflict_table.c.change1_key,
+                                secondaryjoin=change_table.c.key==change_conflict_table.c.change2_key,
+                                ),
+        conflicts2=relationship(Change,
+                                secondary=change_conflict_table,
+                                primaryjoin=change_table.c.key==change_conflict_table.c.change2_key,
+                                secondaryjoin=change_table.c.key==change_conflict_table.c.change1_key,
+                                ),
         revisions=relationship(Revision, backref='change',
                                order_by=revision_table.c.number,
                                cascade='all, delete-orphan'),
@@ -668,20 +765,26 @@ class DatabaseSession(object):
     def vacuum(self):
         self.session().execute("VACUUM")
 
-    def getProjects(self, subscribed=False, unreviewed=False):
+    def getProjects(self, subscribed=False, unreviewed=False, topicless=False):
         """Retrieve projects.
 
         :param subscribed: If True limit to only subscribed projects.
         :param unreviewed: If True limit to only projects with unreviewed
             changes.
+        :param topicless: If True limit to only projects without topics.
         """
         query = self.session().query(Project)
         if subscribed:
             query = query.filter_by(subscribed=subscribed)
             if unreviewed:
                 query = query.filter(exists().where(Project.unreviewed_changes))
+        if topicless:
+            query = query.filter_by(topics=None)
         return query.order_by(Project.name).all()
 
+    def getTopics(self):
+        return self.session().query(Topic).order_by(Topic.sequence).all()
+
     def getProject(self, key):
         try:
             return self.session().query(Project).filter_by(key=key).one()
@@ -694,15 +797,30 @@ class DatabaseSession(object):
         except sqlalchemy.orm.exc.NoResultFound:
             return None
 
+    def getTopic(self, key):
+        try:
+            return self.session().query(Topic).filter_by(key=key).one()
+        except sqlalchemy.orm.exc.NoResultFound:
+            return None
+
+    def getTopicByName(self, name):
+        try:
+            return self.session().query(Topic).filter_by(name=name).one()
+        except sqlalchemy.orm.exc.NoResultFound:
+            return None
+
     def getSyncQueryByName(self, name):
         try:
             return self.session().query(SyncQuery).filter_by(name=name).one()
         except sqlalchemy.orm.exc.NoResultFound:
             return self.createSyncQuery(name)
 
-    def getChange(self, key):
+    def getChange(self, key, lazy=True):
+        query = self.session().query(Change).filter_by(key=key)
+        if not lazy:
+            query = query.options(joinedload(Change.revisions).joinedload(Revision.files).joinedload(File.comments))
         try:
-            return self.session().query(Change).filter_by(key=key).one()
+            return query.one()
         except sqlalchemy.orm.exc.NoResultFound:
             return None
 
@@ -717,12 +835,13 @@ class DatabaseSession(object):
         # the set of supplied IDs. This is used when sync'ing the changesets
         # locally with the remote changes.
         if not ids:
-            return set([])
-        return set([r[0] for r in self.session().query(Change.id).filter(Change.id.in_(ids)).all()])
+            return set()
+        query = self.session().query(Change.id)
+        return set(ids).intersection(r[0] for r in query.all())
 
-    def getChangeByChangeID(self, change_id):
+    def getChangesByChangeID(self, change_id):
         try:
-            return self.session().query(Change).filter_by(change_id=change_id).one()
+            return self.session().query(Change).filter_by(change_id=change_id)
         except sqlalchemy.orm.exc.NoResultFound:
             return None
 
@@ -739,12 +858,14 @@ class DatabaseSession(object):
             return None
 
     def getChanges(self, query, unreviewed=False, sort_by='number'):
-        self.database.log.debug("Search query: %s" % query)
+        self.database.log.debug("Search query: %s sort: %s" % (query, sort_by))
         q = self.session().query(Change).filter(self.search.parse(query))
         if unreviewed:
             q = q.filter(change_table.c.hidden==False, change_table.c.reviewed==False)
         if sort_by == 'updated':
             q = q.order_by(change_table.c.updated)
+        elif sort_by == 'last-seen':
+            q = q.order_by(change_table.c.last_seen)
         else:
             q = q.order_by(change_table.c.number)
         self.database.log.debug("Search SQL: %s" % q)
@@ -766,7 +887,7 @@ class DatabaseSession(object):
             return None
 
     def getRevisionsByParent(self, parent):
-        if isinstance(parent, basestring):
+        if isinstance(parent, six.string_types):
             parent = (parent,)
         try:
             return self.session().query(Revision).filter(Revision.parent.in_(parent)).all()
@@ -872,3 +993,9 @@ class DatabaseSession(object):
         self.session().add(o)
         self.session().flush()
         return o
+
+    def createTopic(self, *args, **kw):
+        o = Topic(*args, **kw)
+        self.session().add(o)
+        self.session().flush()
+        return o
