@@ -12,6 +12,11 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+# Test changes:
+# https://review.openstack.org/275862
+# https://review.openstack.org/119302
+# https://review.openstack.org/133550
+
 import datetime
 import logging
 import difflib
@@ -21,6 +26,7 @@ import re
 
 import git
 import gitdb
+import six
 
 OLD = 0
 NEW = 1
@@ -83,14 +89,14 @@ class CommitContext(object):
             commit.authored_date, commit.author_tz_offset)
         commit_date = self.decorateGitTime(
             commit.committed_date, commit.committer_tz_offset)
-        if type(author.email) is unicode:
+        if isinstance(author.email, six.text_type):
             author_email = author.email
         else:
-            author_email = unicode(author.email, 'utf8')
-        if type(committer.email) is unicode:
+            author_email = author.email.decode('utf8')
+        if isinstance(committer.email, six.text_type):
             committer_email = committer.email
         else:
-            committer_email = unicode(committer.email, 'utf8')
+            committer_email = committer.email.decode('utf8')
         return [u"Parent: %s\n" % parentsha,
                 u"Author: %s <%s>\n" % (author.name, author_email),
                 u"AuthorDate: %s\n" % author_date,
@@ -181,8 +187,9 @@ class DiffFile(object):
     def finalize(self):
         if not self.current_chunk:
             return
-        self.current_chunk.lines = zip(self.current_chunk.oldlines,
-                                       self.current_chunk.newlines)
+        self.current_chunk.lines = list(
+                six.moves.zip(self.current_chunk.oldlines,
+                              self.current_chunk.newlines))
         if not self.chunks:
             self.current_chunk.first = True
         else:
@@ -258,6 +265,8 @@ class Repo(object):
             repo.commit(sha)
         except gitdb.exc.BadObject:
             return False
+        except ValueError:
+            return False
         return True
 
     def fetch(self, url, refspec):
@@ -381,18 +390,12 @@ class Repo(object):
             extra_contexts.append(CommitContext(None, newc))
         contexts = itertools.chain(
             extra_contexts, oldc.diff(
-                newc, color='never',create_patch=True, U=context))
+                newc, color='never', create_patch=True, unified=context))
         for diff_context in contexts:
             # Each iteration of this is a file
             f = DiffFile()
-            if diff_context.a_blob:
-                f.oldname = diff_context.a_blob.path
-            if diff_context.b_blob:
-                f.newname = diff_context.b_blob.path
-            # TODO(jeblair): if/when https://github.com/gitpython-developers/GitPython/pull/266 merges,
-            # remove above 4 lines and replace with these two:
-            # f.oldname = diff_context.a_path
-            # f.newname = diff_context.b_path
+            f.oldname = diff_context.a_path
+            f.newname = diff_context.b_path
             if diff_context.new_file:
                 f.oldname = 'Empty file'
                 f.old_empty = True
@@ -407,7 +410,11 @@ class Repo(object):
             oldchunk = []
             newchunk = []
             prev_key = ''
-            diff_lines = diff_context.diff.split('\n')
+            if isinstance(diff_context.diff, six.string_types):
+                diff_text = diff_context.diff
+            else:
+                diff_text = diff_context.diff.decode('utf-8')
+            diff_lines = diff_text.split('\n')
             for i, line in enumerate(diff_lines):
                 last_line = (i == len(diff_lines)-1)
                 if line.startswith('---'):
@@ -512,4 +519,4 @@ def get_repo(project_name, config):
     local_path = os.path.join(config.git_root, project_name)
     local_root = os.path.abspath(config.git_root)
     assert os.path.commonprefix((local_root, local_path)) == local_root
-    return Repo(config.url+'p/'+project_name, local_path)
+    return Repo(config.git_url + project_name, local_path)
