@@ -29,6 +29,7 @@ DEFAULT_PALETTE={
     'focused-min-label': ['light red,standout', ''],
     'link': ['dark blue', ''],
     'focused-link': ['light blue', ''],
+    'footer': ['light gray', 'dark gray'],
     # Diff
     'context-button': ['dark magenta', ''],
     'focused-context-button': ['light magenta', ''],
@@ -75,9 +76,11 @@ DEFAULT_PALETTE={
     'unreviewed-project': ['white', ''],
     'subscribed-project': ['default', ''],
     'unsubscribed-project': ['dark gray', ''],
+    'marked-project': ['light cyan', ''],
     'focused-unreviewed-project': ['white,standout', ''],
     'focused-subscribed-project': ['default,standout', ''],
     'focused-unsubscribed-project': ['dark gray,standout', ''],
+    'focused-marked-project': ['light cyan,standout', ''],
     # change list
     'unreviewed-change': ['default', ''],
     'reviewed-change': ['dark gray', ''],
