@@ -20,8 +20,24 @@ from sqlalchemy.sql.expression import and_, or_, not_, select, func
 
 import gertty.db
 import gertty.search
+from gertty.search.tokenizer import tokens  # NOQA
 
-from tokenizer import tokens  # NOQA
+def age_to_delta(delta, unit):
+    if unit in ['seconds', 'second', 'sec', 's']:
+        pass
+    elif unit in ['minutes', 'minute', 'min', 'm']:
+        delta = delta * 60
+    elif unit in ['hours', 'hour', 'hr', 'h']:
+        delta = delta * 60 * 60
+    elif unit in ['days', 'day', 'd']:
+        delta = delta * 60 * 60 * 24
+    elif unit in ['weeks', 'week', 'w']:
+        delta = delta * 60 * 60 * 24 * 7
+    elif unit in ['months', 'month', 'mon']:
+        delta = delta * 60 * 60 * 24 * 30
+    elif unit in ['years', 'year', 'y']:
+        delta = delta * 60 * 60 * 24 * 365
+    return delta
 
 def SearchParser():
     precedence = (  # NOQA
@@ -61,11 +77,13 @@ def SearchParser():
 
     def p_term(p):
         '''term : age_term
+                | recentlyseen_term
                 | change_term
                 | owner_term
                 | reviewer_term
                 | commit_term
                 | project_term
+                | projects_term
                 | project_key_term
                 | branch_term
                 | topic_term
@@ -92,22 +110,20 @@ def SearchParser():
         now = datetime.datetime.utcnow()
         delta = p[2]
         unit = p[3]
-        if unit in ['seconds', 'second', 'sec', 's']:
-            pass
-        elif unit in ['minutes', 'minute', 'min', 'm']:
-            delta = delta * 60
-        elif unit in ['hours', 'hour', 'hr', 'h']:
-            delta = delta * 60 * 60
-        elif unit in ['days', 'day', 'd']:
-            delta = delta * 60 * 60 * 24
-        elif unit in ['weeks', 'week', 'w']:
-            delta = delta * 60 * 60 * 24 * 7
-        elif unit in ['months', 'month', 'mon']:
-            delta = delta * 60 * 60 * 24 * 30
-        elif unit in ['years', 'year', 'y']:
-            delta = delta * 60 * 60 * 24 * 365
+        delta = age_to_delta(delta, unit)
         p[0] = gertty.db.change_table.c.updated < (now-datetime.timedelta(seconds=delta))
 
+    def p_recentlyseen_term(p):
+        '''recentlyseen_term : OP_RECENTLYSEEN NUMBER string'''
+        # A gertty extension
+        now = datetime.datetime.utcnow()
+        delta = p[2]
+        unit = p[3]
+        delta = age_to_delta(delta, unit)
+        s = select([func.datetime(func.max(gertty.db.change_table.c.last_seen), '-%s seconds' % delta)],
+                   correlate=False)
+        p[0] = gertty.db.change_table.c.last_seen >= s
+
     def p_change_term(p):
         '''change_term : OP_CHANGE CHANGE_ID
                        | OP_CHANGE NUMBER'''
@@ -163,6 +179,10 @@ def SearchParser():
         else:
             p[0] = gertty.db.project_table.c.name == p[2]
 
+    def p_projects_term(p):
+        '''projects_term : OP_PROJECTS string'''
+        p[0] = gertty.db.project_table.c.name.like('%s%%' % p[2])
+
     def p_project_key_term(p):
         '''project_key_term : OP_PROJECT_KEY NUMBER'''
         p[0] = gertty.db.change_table.c.project_key == p[2]
