@@ -20,11 +20,8 @@ import math
 import os
 import re
 import threading
-import urllib
-import urlparse
 import json
 import time
-import Queue
 import datetime
 
 import dateutil.parser
@@ -34,9 +31,13 @@ except:
     pass
 import requests
 import requests.utils
+import six
+from six.moves import queue
+from six.moves.urllib import parse as urlparse
 
 import gertty.version
 from gertty import gitrepo
+from gertty.auth import FormAuth
 
 HIGH_PRIORITY=0
 NORMAL_PRIORITY=1
@@ -46,6 +47,8 @@ TIMEOUT=30
 
 CLOSED_STATUSES = ['MERGED', 'ABANDONED']
 
+class OfflineError(Exception):
+    pass
 
 class MultiQueue(object):
     def __init__(self, priorities):
@@ -56,12 +59,17 @@ class MultiQueue(object):
         for key in priorities:
             self.queues[key] = collections.deque()
         self.condition = threading.Condition()
+        self.incomplete = []
 
     def qsize(self):
         count = 0
-        for queue in self.queues.values():
-            count += len(queue)
-        return count
+        self.condition.acquire()
+        try:
+            for queue in self.queues.values():
+                count += len(queue)
+            return count + len(self.incomplete)
+        finally:
+            self.condition.release()
 
     def put(self, item, priority):
         added = False
@@ -82,6 +90,7 @@ class MultiQueue(object):
                 for queue in self.queues.values():
                     try:
                         ret = queue.popleft()
+                        self.incomplete.append(ret)
                         return ret
                     except IndexError:
                         pass
@@ -100,6 +109,14 @@ class MultiQueue(object):
             self.condition.release()
         return results
 
+    def complete(self, item):
+        self.condition.acquire()
+        try:
+            if item in self.incomplete:
+                self.incomplete.remove(item)
+        finally:
+            self.condition.release()
+
 
 class UpdateEvent(object):
     def updateRelatedChanges(self, session, change):
@@ -262,7 +279,7 @@ class SyncProjectBranchesTask(Task):
 
     def run(self, sync):
         app = sync.app
-        remote = sync.get('projects/%s/branches/' % urllib.quote_plus(self.project_name))
+        remote = sync.get('projects/%s/branches/' % urlparse.quote_plus(self.project_name))
         remote_branches = set()
         for x in remote:
             m = self.branch_re.match(x['ref'])
@@ -337,27 +354,7 @@ class SyncProjectTask(Task):
                 else:
                     query += ' status:open'
                 queries.append(query)
-        changes = []
-        sortkey = ''
-        done = False
-        while not done:
-            query = '&'.join(queries)
-            # We don't actually want to limit to 500, but that's the server-side default, and
-            # if we don't specify this, we won't get a _more_changes flag.
-            q = 'changes/?n=500%s&%s' % (sortkey, query)
-            self.log.debug('Query: %s ' % (q,))
-            responses = sync.get(q)
-            if len(queries) == 1:
-                responses = [responses]
-            done = True
-            for batch in responses:
-                changes += batch
-                if batch and '_more_changes' in batch[-1]:
-                    done = False
-                    if '_sortkey' in batch[-1]:
-                        sortkey = '&N=%s' % (batch[-1]['_sortkey'],)
-                    else:
-                        sortkey = '&start=%s' % (len(batch))
+        changes = sync.query(queries)
         change_ids = [c['id'] for c in changes]
         with app.db.getSession() as session:
             # Winnow the list of IDs to only the ones in the local DB.
@@ -425,6 +422,7 @@ class SyncQueriedChangesTask(Task):
         changes = []
         sortkey = ''
         done = False
+        offset = 0
         while not done:
             # We don't actually want to limit to 500, but that's the server-side default, and
             # if we don't specify this, we won't get a _more_changes flag.
@@ -435,8 +433,12 @@ class SyncQueriedChangesTask(Task):
             if batch:
                 changes += batch
                 if '_more_changes' in batch[-1]:
-                    sortkey = '&N=%s' % (batch[-1]['_sortkey'],)
                     done = False
+                    if '_sortkey' in batch[-1]:
+                        sortkey = '&N=%s' % (batch[-1]['_sortkey'],)
+                    else:
+                        offset += len(batch)
+                        sortkey = '&start=%s' % (offset,)
         change_ids = [c['id'] for c in changes]
         with app.db.getSession() as session:
             # Winnow the list of IDs to only the ones in the local DB.
@@ -550,6 +552,8 @@ class SyncChangeTask(Task):
         for remote_commit, remote_revision in remote_change.get('revisions', {}).items():
             remote_comments_data = sync.get('changes/%s/revisions/%s/comments' % (self.change_id, remote_commit))
             remote_revision['_gertty_remote_comments_data'] = remote_comments_data
+        remote_conflicts = sync.query(['q=status:open+is:mergeable+conflicts:%s' %
+                                       remote_change['_number']])
         fetches = collections.defaultdict(list)
         parent_commits = set()
         with app.db.getSession() as session:
@@ -571,6 +575,7 @@ class SyncChangeTask(Task):
                 result = ChangeAddedEvent(change)
             else:
                 result = ChangeUpdatedEvent(change)
+            app.project_cache.clear(change.project)
             self.results.append(result)
             change.owner = account
             if change.status != remote_change['status']:
@@ -583,11 +588,34 @@ class SyncChangeTask(Task):
             change.subject = remote_change['subject']
             change.updated = dateutil.parser.parse(remote_change['updated'])
             change.topic = remote_change.get('topic')
+            unseen_conflicts = [x.id for x in change.conflicts]
+            for remote_conflict in remote_conflicts:
+                conflict_id = remote_conflict['id']
+                conflict = session.getChangeByID(conflict_id)
+                if not conflict:
+                    self.log.info("Need to sync conflicting change %s for change %s.",
+                                  conflict_id, change.number)
+                    sync.submitTask(SyncChangeTask(conflict_id, priority=self.priority))
+                else:
+                    if conflict not in change.conflicts:
+                        self.log.info("Added conflict %s for change %s in local DB.",
+                                      conflict.number, change.number)
+                        change.addConflict(conflict)
+                        self.results.append(ChangeUpdatedEvent(conflict))
+                if conflict_id in unseen_conflicts:
+                    unseen_conflicts.remove(conflict_id)
+            for conflict_id in unseen_conflicts:
+                conflict = session.getChangeByID(conflict_id)
+                self.log.info("Deleted conflict %s for change %s in local DB.",
+                              conflict.number, change.number)
+                change.delConflict(conflict)
+                self.results.append(ChangeUpdatedEvent(conflict))
             repo = gitrepo.get_repo(change.project.name, app.config)
             new_revision = False
             for remote_commit, remote_revision in remote_change.get('revisions', {}).items():
                 revision = session.getRevisionByCommit(remote_commit)
                 # TODO: handle multiple parents
+                url = sync.app.config.git_url + change.project.name
                 if 'anonymous http' in remote_revision['fetch']:
                     ref = remote_revision['fetch']['anonymous http']['ref']
                     url = remote_revision['fetch']['anonymous http']['url']
@@ -597,8 +625,8 @@ class SyncChangeTask(Task):
                     ref = remote_revision['fetch']['http']['ref']
                     url = list(urlparse.urlsplit(sync.app.config.url + change.project.name))
                     url[1] = '%s:%s@%s' % (
-                        urllib.quote_plus(sync.app.config.username),
-                        urllib.quote_plus(sync.app.config.password), url[1])
+                        urlparse.quote_plus(sync.app.config.username),
+                        urlparse.quote_plus(sync.app.config.password), url[1])
                     url = urlparse.urlunsplit(url)
                 elif 'ssh' in remote_revision['fetch']:
                     ref = remote_revision['fetch']['ssh']['ref']
@@ -829,6 +857,7 @@ class SyncChangeTask(Task):
                     if change.reviewed:
                         change.reviewed = False
                         result.review_flag_changed = True
+                        app.project_cache.clear(change.project)
         for url, refs in fetches.items():
             self.log.debug("Fetching from %s with refs %s", url, refs)
             try:
@@ -1249,7 +1278,7 @@ class PruneChangeTask(Task):
                 change.project.name, change_ref))
             try:
                 repo.deleteRef(change_ref)
-            except OSError, e:
+            except OSError as e:
                 if e.errno not in [errno.EISDIR, errno.EPERM]:
                     raise
             session.delete(change)
@@ -1272,7 +1301,7 @@ class VacuumDatabaseTask(Task):
             session.vacuum()
 
 class Sync(object):
-    def __init__(self, app):
+    def __init__(self, app, disable_background_sync):
         self.user_agent = 'Gertty/%s %s' % (gertty.version.version_info.release_string(),
                                             requests.utils.default_user_agent())
         self.version = (0, 0, 0)
@@ -1281,25 +1310,28 @@ class Sync(object):
         self.app = app
         self.log = logging.getLogger('gertty.sync')
         self.queue = MultiQueue([HIGH_PRIORITY, NORMAL_PRIORITY, LOW_PRIORITY])
-        self.result_queue = Queue.Queue()
+        self.result_queue = queue.Queue()
         self.session = requests.Session()
         if self.app.config.auth_type == 'basic':
             authclass = requests.auth.HTTPBasicAuth
+        elif self.app.config.auth_type == 'form':
+            authclass = FormAuth
         else:
             authclass = requests.auth.HTTPDigestAuth
         self.auth = authclass(
             self.app.config.username, self.app.config.password)
         self.submitTask(GetVersionTask(HIGH_PRIORITY))
         self.submitTask(SyncOwnAccountTask(HIGH_PRIORITY))
-        self.submitTask(CheckReposTask(HIGH_PRIORITY))
-        self.submitTask(UploadReviewsTask(HIGH_PRIORITY))
-        self.submitTask(SyncProjectListTask(HIGH_PRIORITY))
-        self.submitTask(SyncSubscribedProjectsTask(NORMAL_PRIORITY))
-        self.submitTask(SyncSubscribedProjectBranchesTask(LOW_PRIORITY))
-        self.submitTask(PruneDatabaseTask(self.app.config.expire_age, LOW_PRIORITY))
-        self.periodic_thread = threading.Thread(target=self.periodicSync)
-        self.periodic_thread.daemon = True
-        self.periodic_thread.start()
+        if not disable_background_sync:
+            self.submitTask(CheckReposTask(HIGH_PRIORITY))
+            self.submitTask(UploadReviewsTask(HIGH_PRIORITY))
+            self.submitTask(SyncProjectListTask(HIGH_PRIORITY))
+            self.submitTask(SyncSubscribedProjectsTask(NORMAL_PRIORITY))
+            self.submitTask(SyncSubscribedProjectBranchesTask(LOW_PRIORITY))
+            self.submitTask(PruneDatabaseTask(self.app.config.expire_age, LOW_PRIORITY))
+            self.periodic_thread = threading.Thread(target=self.periodicSync)
+            self.periodic_thread.daemon = True
+            self.periodic_thread.start()
 
     def periodicSync(self):
         hourly = time.time()
@@ -1333,30 +1365,37 @@ class Sync(object):
         try:
             task.run(self)
             task.complete(True)
-        except requests.ConnectionError, e:
+            self.queue.complete(task)
+        except (requests.ConnectionError, OfflineError) as e:
             self.log.warning("Offline due to: %s" % (e,))
             if not self.offline:
                 self.submitTask(GetVersionTask(HIGH_PRIORITY))
                 self.submitTask(UploadReviewsTask(HIGH_PRIORITY))
             self.offline = True
             self.app.status.update(offline=True, refresh=False)
-            os.write(pipe, 'refresh\n')
+            os.write(pipe, six.b('refresh\n'))
             time.sleep(30)
             return task
         except Exception:
             task.complete(False)
+            self.queue.complete(task)
             self.log.exception('Exception running task %s' % (task,))
             self.app.status.update(error=True, refresh=False)
         self.offline = False
         self.app.status.update(offline=False, refresh=False)
         for r in task.results:
             self.result_queue.put(r)
-        os.write(pipe, 'refresh\n')
+        os.write(pipe, six.b('refresh\n'))
         return None
 
     def url(self, path):
         return self.app.config.url + 'a/' + path
 
+    def checkResponse(self, response):
+        self.log.debug('HTTP status code: %d', response.status_code)
+        if response.status_code == 503:
+            raise OfflineError("Received 503 status code")
+
     def get(self, path):
         url = self.url(path)
         self.log.debug('GET: %s' % (url,))
@@ -1366,6 +1405,7 @@ class Sync(object):
                              headers = {'Accept': 'application/json',
                                         'Accept-Encoding': 'gzip',
                                         'User-Agent': self.user_agent})
+        self.checkResponse(r)
         if r.status_code == 200:
             ret = json.loads(r.text[4:])
             if len(ret):
@@ -1373,8 +1413,6 @@ class Sync(object):
             else:
                 self.log.debug('200 OK, No body.')
             return ret
-        else:
-            self.log.warn('HTTP response: %d', r.status_code)
 
     def post(self, path, data):
         url = self.url(path)
@@ -1385,6 +1423,7 @@ class Sync(object):
                               auth=self.auth, timeout=TIMEOUT,
                               headers = {'Content-Type': 'application/json;charset=UTF-8',
                                          'User-Agent': self.user_agent})
+        self.checkResponse(r)
         self.log.debug('Received: %s' % (r.text,))
         ret = None
         if r.text and len(r.text)>4:
@@ -1404,6 +1443,7 @@ class Sync(object):
                              auth=self.auth, timeout=TIMEOUT,
                              headers = {'Content-Type': 'application/json;charset=UTF-8',
                                         'User-Agent': self.user_agent})
+        self.checkResponse(r)
         self.log.debug('Received: %s' % (r.text,))
 
     def delete(self, path, data):
@@ -1415,6 +1455,7 @@ class Sync(object):
                                 auth=self.auth, timeout=TIMEOUT,
                                 headers = {'Content-Type': 'application/json;charset=UTF-8',
                                            'User-Agent': self.user_agent})
+        self.checkResponse(r)
         self.log.debug('Received: %s' % (r.text,))
 
     def syncSubscribedProjects(self):
@@ -1455,3 +1496,29 @@ class Sync(object):
             micro = int(parts[2])
         self.version = (major, minor, micro)
         self.log.info("Remote version is: %s (parsed as %s)" % (version, self.version))
+
+    def query(self, queries):
+        changes = []
+        sortkey = ''
+        done = False
+        offset = 0
+        while not done:
+            query = '&'.join(queries)
+            # We don't actually want to limit to 500, but that's the server-side default, and
+            # if we don't specify this, we won't get a _more_changes flag.
+            q = 'changes/?n=500%s&%s' % (sortkey, query)
+            self.log.debug('Query: %s' % (q,))
+            responses = self.get(q)
+            if len(queries) == 1:
+                responses = [responses]
+            done = True
+            for batch in responses:
+                changes += batch
+                if batch and '_more_changes' in batch[-1]:
+                    done = False
+                    if '_sortkey' in batch[-1]:
+                        sortkey = '&N=%s' % (batch[-1]['_sortkey'],)
+                    else:
+                        offset += len(batch)
+                        sortkey = '&start=%s' % (offset,)
+        return changes
