@@ -13,10 +13,16 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+import collections
 import datetime
 import logging
-import urlparse
+try:
+    import ordereddict
+except:
+    pass
+import textwrap
 
+from six.moves.urllib import parse as urlparse
 import urwid
 
 from gertty import gitrepo
@@ -28,6 +34,11 @@ from gertty.view import unified_diff as view_unified_diff
 from gertty.view import mouse_scroll_decorator
 import gertty.view
 
+try:
+    OrderedDict = collections.OrderedDict
+except AttributeError:
+    OrderedDict = ordereddict.OrderedDict
+
 class EditTopicDialog(mywid.ButtonDialog):
     signals = ['save', 'cancel']
     def __init__(self, app, topic):
@@ -56,7 +67,7 @@ class EditTopicDialog(mywid.ButtonDialog):
             return None
         return key
 
-class CherryPickDialog(urwid.WidgetWrap):
+class CherryPickDialog(urwid.WidgetWrap, mywid.LineBoxTitlePropertyMixin):
     signals = ['save', 'cancel']
     def __init__(self, app, change):
         save_button = mywid.FixedButton('Propose Change')
@@ -87,9 +98,9 @@ class CherryPickDialog(urwid.WidgetWrap):
         super(CherryPickDialog, self).__init__(urwid.LineBox(fill,
                                                              'Propose Change to Branch'))
 
-class ReviewDialog(urwid.WidgetWrap):
+class ReviewDialog(urwid.WidgetWrap, mywid.LineBoxTitlePropertyMixin):
     signals = ['submit', 'save', 'cancel']
-    def __init__(self, app, revision_key):
+    def __init__(self, app, revision_key, message=''):
         self.revision_key = revision_key
         self.app = app
         save_button = mywid.FixedButton(u'Save')
@@ -107,7 +118,6 @@ class ReviewDialog(urwid.WidgetWrap):
         values = {}
         descriptions = {}
         self.button_groups = {}
-        message = ''
         with self.app.db.getSession() as session:
             revision = session.getRevision(self.revision_key)
             change = revision.change
@@ -132,13 +142,20 @@ class ReviewDialog(urwid.WidgetWrap):
                         values[label.category] = []
                     values[label.category].append(label.value)
                 draft_approvals = {}
-                for approval in change.draft_approvals:
-                    draft_approvals[approval.category] = approval
+                prior_approvals = {}
+                for approval in change.approvals:
+                    if approval.reviewer.username == self.app.config.username:
+                        if approval.draft:
+                            draft_approvals[approval.category] = approval
+                        else:
+                            prior_approvals[approval.category] = approval
                 for category in categories:
                     rows.append(urwid.Text(category))
                     group = []
                     self.button_groups[category] = group
                     current = draft_approvals.get(category)
+                    if current is None:
+                        current = prior_approvals.get(category)
                     if current is None:
                         current = 0
                     else:
@@ -170,7 +187,7 @@ class ReviewDialog(urwid.WidgetWrap):
                 m = revision.getDraftMessage()
             if m:
                 message = m.message
-        self.message = mywid.MyEdit("Message: \n", edit_text=message,
+        self.message = mywid.MyEdit(u"Message: \n", edit_text=message,
                                     multiline=True, ring=app.ring)
         rows.append(self.message)
         rows.append(urwid.Divider())
@@ -206,9 +223,10 @@ class ReviewButton(mywid.FixedButton):
         urwid.connect_signal(self, 'click',
             lambda button: self.openReview())
 
-    def openReview(self):
+    def openReview(self, message=''):
         self.dialog = ReviewDialog(self.change_view.app,
-                                   self.revision_row.revision_key)
+                                   self.revision_row.revision_key,
+                                   message=message)
         urwid.connect_signal(self.dialog, 'save',
             lambda button: self.closeReview(True, False))
         urwid.connect_signal(self.dialog, 'submit',
@@ -319,7 +337,7 @@ class RevisionRow(urwid.WidgetWrap):
     def cherryPick(self, button):
         self.app.localCherryPickCommit(self.project_name, self.commit_sha)
 
-class ChangeButton(mywid.FixedButton):
+class ChangeButton(urwid.Button):
     button_left = urwid.Text(u' ')
     button_right = urwid.Text(u' ')
 
@@ -341,13 +359,54 @@ class ChangeButton(mywid.FixedButton):
             self.change_view.app.error(e.message)
 
 class ChangeMessageBox(mywid.HyperText):
-    def __init__(self, app, message):
+    def __init__(self, change_view, message):
         super(ChangeMessageBox, self).__init__(u'')
-        self.app = app
+        self.change_view = change_view
+        self.app = change_view.app
         self.refresh(message)
 
+    def formatReply(self):
+        text = self.message_text
+        pgraphs = []
+        pgraph_accumulator = []
+        wrap = True
+        for line in text.split('\n')[2:]:
+            if line.startswith('> '):
+                wrap = False
+                line = '> ' + line
+            if not line:
+                if pgraph_accumulator:
+                    pgraphs.append((wrap, '\n'.join(pgraph_accumulator)))
+                    pgraph_accumulator = []
+                    wrap = True
+                continue
+            pgraph_accumulator.append(line)
+        if pgraph_accumulator:
+            pgraphs.append((wrap, '\n'.join(pgraph_accumulator)))
+            pgraph_accumulator = []
+            wrap = True
+        wrapper = textwrap.TextWrapper(initial_indent='> ',
+                                       subsequent_indent='> ')
+        wrapped_pgraphs = []
+        for wrap, pgraph in pgraphs:
+            if wrap:
+                wrapped_pgraphs.append('\n'.join(wrapper.wrap(pgraph)))
+            else:
+                wrapped_pgraphs.append(pgraph)
+        return '\n>\n'.join(wrapped_pgraphs)
+
+    def reply(self):
+        reply_text = self.formatReply()
+        if reply_text:
+            reply_text = self.message_author + ' wrote:\n\n' + reply_text + '\n'
+        row = self.change_view.revision_rows[self.revision_key]
+        row.review_button.openReview(reply_text)
+
     def refresh(self, message):
+        self.revision_key = message.revision.key
         self.message_created = message.created
+        self.message_author = message.author_name
+        self.message_text = message.message
         created = self.app.time(message.created)
         lines = message.message.split('\n')
         if message.draft:
@@ -356,15 +415,32 @@ class ChangeMessageBox(mywid.HyperText):
         if message.author.username == self.app.config.username:
             name_style = 'change-message-own-name'
             header_style = 'change-message-own-header'
+            reviewer_string = message.author.name
         else:
             name_style = 'change-message-name'
             header_style = 'change-message-header'
-        text = [(name_style, message.author_name),
+            if message.author.email:
+                reviewer_string = "%s <%s>" % (
+                    message.author.name,
+                    message.author.email)
+            else:
+                reviewer_string = message.author.name
+
+        text = [(name_style, reviewer_string),
                 (header_style, ': '+lines.pop(0)),
                 (header_style,
                  created.strftime(' (%Y-%m-%d %H:%M:%S%z)'))]
         if message.draft and not message.pending:
             text.append(('change-message-draft', ' (draft)'))
+        else:
+            link = mywid.Link('< Reply >',
+                              'revision-button',
+                              'focused-revision-button')
+            urwid.connect_signal(link, 'selected',
+                                 lambda link:self.reply())
+            text.append(' ')
+            text.append(link)
+
         if lines and lines[-1]:
             lines.append('')
         comment_text = ['\n'.join(lines)]
@@ -385,55 +461,57 @@ class CommitMessageBox(mywid.HyperText):
 
 @mouse_scroll_decorator.ScrollByWheel
 class ChangeView(urwid.WidgetWrap):
-    def help(self):
-        key = self.app.config.keymap.formatKeys
-        ret = [
-            (key(keymap.LOCAL_CHECKOUT),
+    def getCommands(self):
+        return [
+            (keymap.LOCAL_CHECKOUT,
              "Checkout the most recent revision into the local repo"),
-            (key(keymap.DIFF),
+            (keymap.DIFF,
              "Show the diff of the most recent revision"),
-            (key(keymap.TOGGLE_HIDDEN),
+            (keymap.TOGGLE_HIDDEN,
              "Toggle the hidden flag for the current change"),
-            (key(keymap.NEXT_CHANGE),
+            (keymap.NEXT_CHANGE,
              "Go to the next change in the list"),
-            (key(keymap.PREV_CHANGE),
+            (keymap.PREV_CHANGE,
              "Go to the previous change in the list"),
-            (key(keymap.REVIEW),
+            (keymap.REVIEW,
              "Leave a review for the most recent revision"),
-            (key(keymap.TOGGLE_HELD),
+            (keymap.TOGGLE_HELD,
              "Toggle the held flag for the current change"),
-            (key(keymap.TOGGLE_HIDDEN_COMMENTS),
+            (keymap.TOGGLE_HIDDEN_COMMENTS,
              "Toggle display of hidden comments"),
-            (key(keymap.SEARCH_RESULTS),
+            (keymap.SEARCH_RESULTS,
              "Back to the list of changes"),
-            (key(keymap.TOGGLE_REVIEWED),
+            (keymap.TOGGLE_REVIEWED,
              "Toggle the reviewed flag for the current change"),
-            (key(keymap.TOGGLE_STARRED),
+            (keymap.TOGGLE_STARRED,
              "Toggle the starred flag for the current change"),
-            (key(keymap.LOCAL_CHERRY_PICK),
+            (keymap.LOCAL_CHERRY_PICK,
              "Cherry-pick the most recent revision onto the local repo"),
-            (key(keymap.ABANDON_CHANGE),
+            (keymap.ABANDON_CHANGE,
              "Abandon this change"),
-            (key(keymap.EDIT_COMMIT_MESSAGE),
+            (keymap.EDIT_COMMIT_MESSAGE,
              "Edit the commit message of this change"),
-            (key(keymap.REBASE_CHANGE),
+            (keymap.REBASE_CHANGE,
              "Rebase this change (remotely)"),
-            (key(keymap.RESTORE_CHANGE),
+            (keymap.RESTORE_CHANGE,
              "Restore this change"),
-            (key(keymap.REFRESH),
+            (keymap.REFRESH,
              "Refresh this change"),
-            (key(keymap.EDIT_TOPIC),
+            (keymap.EDIT_TOPIC,
              "Edit the topic of this change"),
-            (key(keymap.SUBMIT_CHANGE),
+            (keymap.SUBMIT_CHANGE,
              "Submit this change"),
-            (key(keymap.CHERRY_PICK_CHANGE),
+            (keymap.CHERRY_PICK_CHANGE,
              "Propose this change to another branch"),
             ]
 
+    def help(self):
+        key = self.app.config.keymap.formatKeys
+        commands = self.getCommands()
+        ret = [(c[0], key(c[0]), c[1]) for c in commands]
         for k in self.app.config.reviewkeys.values():
             action = ', '.join(['{category}:{value}'.format(**a) for a in k['approvals']])
-            ret.append((keymap.formatKey(k['key']), action))
-
+            ret.append(('', keymap.formatKey(k['key']), action))
         return ret
 
     def __init__(self, app, change_key):
@@ -445,6 +523,7 @@ class ChangeView(urwid.WidgetWrap):
         self.message_rows = {}
         self.last_revision_key = None
         self.hide_comments = True
+        self.marked_seen = False
         self.change_id_label = mywid.TextButton(u'', on_press=self.searchChangeId)
         self.owner_label = mywid.TextButton(u'', on_press=self.searchOwner)
         self.project_label = mywid.TextButton(u'', on_press=self.searchProject)
@@ -485,7 +564,9 @@ class ChangeView(urwid.WidgetWrap):
         self.depends_on_rows = {}
         self.needed_by = urwid.Pile([])
         self.needed_by_rows = {}
-        self.related_changes = urwid.Pile([self.depends_on, self.needed_by])
+        self.conflicts_with = urwid.Pile([])
+        self.conflicts_with_rows = {}
+        self.related_changes = urwid.Pile([self.depends_on, self.needed_by, self.conflicts_with])
         self.results = mywid.HyperText(u'') # because it scrolls better than a table
         self.grid = mywid.MyGridFlow([change_info, self.commit_message, votes, self.results],
                                      cell_width=80, h_sep=2, v_sep=1, align='left')
@@ -547,7 +628,12 @@ class ChangeView(urwid.WidgetWrap):
 
     def refresh(self):
         with self.app.db.getSession() as session:
-            change = session.getChange(self.change_key)
+            change = session.getChange(self.change_key, lazy=False)
+            # When we first open the change, update its last_seen
+            # time.
+            if not self.marked_seen:
+                change.last_seen = datetime.datetime.utcnow()
+                self.marked_seen = True
             self.topic = change.topic or ''
             self.pending_status_message = change.pending_status_message or ''
             reviewed = hidden = starred = held = ''
@@ -572,7 +658,12 @@ class ChangeView(urwid.WidgetWrap):
                 self.owner_email = None
 
             self.change_id_label.text.set_text(('change-data', change.change_id))
-            self.owner_label.text.set_text(('change-data', change.owner_name))
+            if change.owner.email:
+                owner_string = '%s <%s>' % (change.owner_name,
+                                            change.owner.email)
+            else:
+                owner_string = change.owner_name
+            self.owner_label.text.set_text(('change-data', owner_string))
             self.project_label.text.set_text(('change-data', change.project.name))
             self.branch_label.set_text(('change-data', change.branch))
             self.topic_label.text.set_text(('change-data', self.topic))
@@ -665,7 +756,8 @@ class ChangeView(urwid.WidgetWrap):
                     for commentlink in self.app.config.commentlinks:
                         results = commentlink.getTestResults(self.app, message.message)
                         if results:
-                            result_system = result_systems.get(message.author.name, {})
+                            result_system = result_systems.get(message.author.name,
+                                                               OrderedDict())
                             result_systems[message.author.name] = result_system
                             result_system.update(results)
                 skip = False
@@ -682,7 +774,7 @@ class ChangeView(urwid.WidgetWrap):
             for message in display_messages:
                 row = self.message_rows.get(message.key)
                 if not row:
-                    box = ChangeMessageBox(self.app, message)
+                    box = ChangeMessageBox(self, message)
                     row = urwid.Padding(box, width=80)
                     self.listbox.body.insert(listbox_index, row)
                     self.message_rows[message.key] = row
@@ -752,6 +844,8 @@ class ChangeView(urwid.WidgetWrap):
             subject = parent.change.subject
             if parent != parent.change.revisions[-1]:
                 subject += ' [OUTDATED]'
+            if parent.change.status == 'ABANDONED':
+                subject += ' [ABANDONED]'
             parents[parent.change.key] = subject
         self._updateDependenciesWidget(parents,
                                        self.depends_on, self.depends_on_rows,
@@ -762,21 +856,34 @@ class ChangeView(urwid.WidgetWrap):
         children.update((r.change.key, r.change.subject)
                         for r in session.getRevisionsByParent([revision.commit for revision in change.revisions])
                         if (r.change.status != 'MERGED' and
+                            r.change.status != 'ABANDONED' and
                             r == r.change.revisions[-1]))
         self._updateDependenciesWidget(children,
                                        self.needed_by, self.needed_by_rows,
                                        header='Needed by:')
 
+        # Handle conflicts_with
+        conflicts = {}
+        conflicts.update((c.key, c.subject)
+                        for c in change.conflicts
+                        if (c.status != 'MERGED' and
+                            c.status != 'ABANDONED'))
+        self._updateDependenciesWidget(conflicts,
+                                       self.conflicts_with, self.conflicts_with_rows,
+                                       header='Conflicts with:')
+
 
     def toggleReviewed(self):
         with self.app.db.getSession() as session:
             change = session.getChange(self.change_key)
             change.reviewed = not change.reviewed
+            self.app.project_cache.clear(change.project)
 
     def toggleHidden(self):
         with self.app.db.getSession() as session:
             change = session.getChange(self.change_key)
             change.hidden = not change.hidden
+            self.app.project_cache.clear(change.project)
 
     def toggleStarred(self):
         with self.app.db.getSession() as session:
