@@ -15,15 +15,41 @@
 
 import datetime
 import logging
+
+import six
 import urwid
 
 from gertty import keymap
+from gertty import mywid
 from gertty import sync
 from gertty.view import change as view_change
 from gertty.view import mouse_scroll_decorator
 import gertty.view
 
 
+class ColumnInfo(object):
+    def __init__(self, name, packing, value):
+        self.name = name
+        self.packing = packing
+        self.value = value
+        self.options = (packing, value)
+        if packing == 'given':
+            self.spacing = value + 1
+        else:
+            self.spacing = (value * 8) + 1
+
+
+COLUMNS = [
+    ColumnInfo('Number',  'given',   6),
+    ColumnInfo('Subject', 'weight',  4),
+    ColumnInfo('Project', 'weight',  1),
+    ColumnInfo('Branch',  'weight',  1),
+    ColumnInfo('Topic',   'weight',  1),
+    ColumnInfo('Owner',   'weight',  1),
+    ColumnInfo('Updated', 'given',  10),
+]
+
+
 class ThreadStack(object):
     def __init__(self):
         self.stack = []
@@ -46,7 +72,23 @@ class ThreadStack(object):
         return [len(x[1]) for x in self.stack]
 
 
-class ChangeRow(urwid.Button):
+class ChangeListColumns(object):
+    def updateColumns(self):
+        del self.columns.contents[:]
+        cols = self.columns.contents
+        options = self.columns.options
+
+        for colinfo in COLUMNS:
+            if colinfo.name in self.enabled_columns:
+                attr = colinfo.name.lower().replace(' ', '_')
+                cols.append((getattr(self, attr),
+                             options(*colinfo.options)))
+
+        for c in self.category_columns:
+            cols.append(c)
+
+
+class ChangeRow(urwid.Button, ChangeListColumns):
     change_focus_map = {None: 'focused',
                         'unreviewed-change': 'focused-unreviewed-change',
                         'reviewed-change': 'focused-reviewed-change',
@@ -62,31 +104,44 @@ class ChangeRow(urwid.Button):
     def selectable(self):
         return True
 
-    def __init__(self, app, change, prefix, categories, project=False,
-                 owner=False, updated=False, callback=None):
+    def __init__(self, app, change, prefix, categories,
+                 enabled_columns, callback=None):
         super(ChangeRow, self).__init__('', on_press=callback, user_data=change.key)
         self.app = app
         self.change_key = change.key
         self.prefix = prefix
-        self.subject = urwid.Text(u'', wrap='clip')
-        self.number = urwid.Text(u'')
-        self.updated = urwid.Text(u'')
-        self.project = urwid.Text(u'', wrap='clip')
-        self.owner = urwid.Text(u'', wrap='clip')
+        self.enabled_columns = enabled_columns
+        self.subject = mywid.SearchableText(u'', wrap='clip')
+        self.number = mywid.SearchableText(u'')
+        self.updated = mywid.SearchableText(u'')
+        self.project = mywid.SearchableText(u'', wrap='clip')
+        self.owner = mywid.SearchableText(u'', wrap='clip')
+        self.branch = mywid.SearchableText(u'', wrap='clip')
+        self.topic = mywid.SearchableText(u'', wrap='clip')
         self.mark = False
-        cols = [(6, self.number), ('weight', 4, self.subject)]
-        if project:
-            cols.append(('weight', 1, self.project))
-        if owner:
-            cols.append(('weight', 2, self.owner))
-        if updated:
-            cols.append(('fixed', 10, self.updated))
-        self.num_columns = len(cols)
-        self.columns = urwid.Columns(cols, dividechars=1)
+        self.columns = urwid.Columns([], dividechars=1)
         self.row_style = urwid.AttrMap(self.columns, '')
         self._w = urwid.AttrMap(self.row_style, None, focus_map=self.change_focus_map)
+        self.category_columns = []
         self.update(change, categories)
 
+    def search(self, search, attribute):
+        if self.subject.search(search, attribute):
+            return True
+        if self.number.search(search, attribute):
+            return True
+        if self.project.search(search, attribute):
+            return True
+        if self.branch.search(search, attribute):
+            return True
+        if self.owner.search(search, attribute):
+            return True
+        if self.topic.search(search, attribute):
+            return True
+        if self.updated.search(search, attribute):
+            return True
+        return False
+
     def update(self, change, categories):
         if change.reviewed or change.hidden:
             style = 'reviewed-change'
@@ -109,6 +164,8 @@ class ChangeRow(urwid.Button):
         self.number.set_text(str(change.number))
         self.project.set_text(change.project.name.split('/')[-1])
         self.owner.set_text(change.owner_name)
+        self.branch.set_text(change.branch or '')
+        self.topic.set_text(change.topic or '')
         self.project_name = change.project.name
         self.commit_sha = change.revisions[-1].commit
         self.current_revision_key = change.revisions[-1].key
@@ -118,7 +175,8 @@ class ChangeRow(urwid.Button):
             self.updated.set_text(updated_time.strftime("%I:%M %p").upper())
         else:
             self.updated.set_text(updated_time.strftime("%Y-%m-%d"))
-        del self.columns.contents[self.num_columns:]
+
+        self.category_columns = []
         for category in categories:
             v = change.getMaxForCategory(category)
             cat_min, cat_max = change.getMinMaxPermittedForCategory(category)
@@ -136,84 +194,118 @@ class ChangeRow(urwid.Button):
                     val = ('min-label', val)
                 else:
                     val = ('negative-label', val)
-            self.columns.contents.append((urwid.Text(val), self.columns.options('given', 2)))
+            self.category_columns.append((urwid.Text(val),
+                                          self.columns.options('given', 2)))
+        self.updateColumns()
 
-class ChangeListHeader(urwid.WidgetWrap):
-    def __init__(self, project=False, owner=False, updated=False):
-        cols = [(6, urwid.Text(u'Number')), ('weight', 4, urwid.Text(u' Subject'))]
-        if project:
-            cols.append(('weight', 1, urwid.Text(u'Project')))
-        if owner:
-            cols.append(('weight', 2, urwid.Text(u'Owner')))
-        if updated:
-            cols.append(('fixed', 10, urwid.Text(u'Updated')))
-        self.num_columns = len(cols)
-        super(ChangeListHeader, self).__init__(urwid.Columns(cols, dividechars=1))
+class ChangeListHeader(urwid.WidgetWrap, ChangeListColumns):
+    def __init__(self, enabled_columns):
+        self.enabled_columns = enabled_columns
+        self.subject = urwid.Text(u'Subject', wrap='clip')
+        self.number = urwid.Text(u'Number')
+        self.updated = urwid.Text(u'Updated')
+        self.project = urwid.Text(u'Project', wrap='clip')
+        self.owner = urwid.Text(u'Owner', wrap='clip')
+        self.branch = urwid.Text(u'Branch', wrap='clip')
+        self.topic = urwid.Text(u'Topic', wrap='clip')
+        self.columns = urwid.Columns([], dividechars=1)
+        self.category_columns = []
+        super(ChangeListHeader, self).__init__(self.columns)
 
     def update(self, categories):
-        del self._w.contents[self.num_columns:]
+        self.category_columns = []
         for category in categories:
-            self._w.contents.append((urwid.Text(' %s' % category[0]), self._w.options('given', 2)))
+            self.category_columns.append((urwid.Text(' %s' % category[0]),
+                                          self._w.options('given', 2)))
+        self.updateColumns()
+
 
 @mouse_scroll_decorator.ScrollByWheel
-class ChangeListView(urwid.WidgetWrap):
-    def help(self):
-        key = self.app.config.keymap.formatKeys
+class ChangeListView(urwid.WidgetWrap, mywid.Searchable):
+    required_columns = set(['Number', 'Subject', 'Updated'])
+    optional_columns = set(['Topic', 'Branch'])
+
+    def getCommands(self):
         if self.project_key:
             refresh_help = "Sync current project"
         else:
             refresh_help = "Sync subscribed projects"
         return [
-            (key(keymap.TOGGLE_HELD),
+            (keymap.TOGGLE_HELD,
              "Toggle the held flag for the currently selected change"),
-            (key(keymap.LOCAL_CHECKOUT),
+            (keymap.LOCAL_CHECKOUT,
              "Checkout the most recent revision of the selected change into the local repo"),
-            (key(keymap.TOGGLE_HIDDEN),
+            (keymap.TOGGLE_HIDDEN,
              "Toggle the hidden flag for the currently selected change"),
-            (key(keymap.TOGGLE_LIST_REVIEWED),
+            (keymap.TOGGLE_LIST_REVIEWED,
              "Toggle whether only unreviewed or all changes are displayed"),
-            (key(keymap.TOGGLE_REVIEWED),
+            (keymap.TOGGLE_REVIEWED,
              "Toggle the reviewed flag for the currently selected change"),
-            (key(keymap.TOGGLE_STARRED),
+            (keymap.TOGGLE_STARRED,
              "Toggle the starred flag for the currently selected change"),
-            (key(keymap.TOGGLE_MARK),
+            (keymap.TOGGLE_MARK,
              "Toggle the process mark for the currently selected change"),
-            (key(keymap.REFINE_CHANGE_SEARCH),
+            (keymap.REFINE_CHANGE_SEARCH,
              "Refine the current search query"),
-            (key(keymap.EDIT_TOPIC),
+            (keymap.ABANDON_CHANGE,
+             "Abandon the marked changes"),
+            (keymap.EDIT_TOPIC,
              "Set the topic of the marked changes"),
-            (key(keymap.REFRESH),
+            (keymap.RESTORE_CHANGE,
+             "Restore the marked changes"),
+            (keymap.REFRESH,
              refresh_help),
-            (key(keymap.REVIEW),
+            (keymap.REVIEW,
              "Leave reviews for the marked changes"),
-            (key(keymap.SORT_BY_NUMBER),
+            (keymap.SORT_BY_NUMBER,
              "Sort changes by number"),
-            (key(keymap.SORT_BY_UPDATED),
+            (keymap.SORT_BY_UPDATED,
              "Sort changes by how recently the change was updated"),
-            (key(keymap.SORT_BY_REVERSE),
+            (keymap.SORT_BY_REVERSE,
              "Reverse the sort"),
-            (key(keymap.LOCAL_CHERRY_PICK),
+            (keymap.LOCAL_CHERRY_PICK,
              "Cherry-pick the most recent revision of the selected change onto the local repo"),
+            (keymap.INTERACTIVE_SEARCH,
+             "Interactive search"),
             ]
 
+    def help(self):
+        key = self.app.config.keymap.formatKeys
+        commands = self.getCommands()
+        return [(c[0], key(c[0]), c[1]) for c in commands]
+
     def __init__(self, app, query, query_desc=None, project_key=None,
-                 unreviewed=False):
+                 unreviewed=False, sort_by=None, reverse=None):
         super(ChangeListView, self).__init__(urwid.Pile([]))
         self.log = logging.getLogger('gertty.view.change_list')
+        self.searchInit()
         self.app = app
         self.query = query
         self.query_desc = query_desc or query
         self.unreviewed = unreviewed
         self.change_rows = {}
+        self.enabled_columns = set()
+        for colinfo in COLUMNS:
+            if (colinfo.name in self.required_columns or
+                colinfo.name not in self.optional_columns):
+                self.enabled_columns.add(colinfo.name)
+        self.disabled_columns = set()
         self.listbox = urwid.ListBox(urwid.SimpleFocusListWalker([]))
-        self.display_owner = self.display_project = self.display_updated = True
         self.project_key = project_key
-        if project_key is not None:
-            self.display_project = False
-        self.sort_by = app.config.change_list_options['sort-by']
-        self.reverse = app.config.change_list_options['reverse']
-        self.header = ChangeListHeader(self.display_project, self.display_owner,
-                                       self.display_updated)
+        if 'Project' not in self.required_columns and project_key is not None:
+            self.enabled_columns.discard('Project')
+            self.disabled_columns.add('Project')
+        if 'Owner' not in self.required_columns and 'owner:' in query:
+            # This could be or'd with something else, but probably
+            # not.
+            self.enabled_columns.discard('Owner')
+            self.disabled_columns.add('Owner')
+        self.sort_by = sort_by or app.config.change_list_options['sort-by']
+        if reverse is not None:
+            self.reverse = reverse
+        else:
+            self.reverse = app.config.change_list_options['reverse']
+        self.header = ChangeListHeader(self.enabled_columns)
         self.categories = []
         self.refresh()
         self._w.contents.append((app.header, ('pack', 1)))
@@ -246,11 +338,17 @@ class ChangeListView(urwid.WidgetWrap):
                 self.title = u'Unreviewed changes in %s' % self.query_desc
             else:
                 self.title = u'All changes in %s' % self.query_desc
+            self.short_title = self.query_desc
+            if '/' in self.short_title and ' ' not in self.short_title:
+                i = self.short_title.rfind('/')
+                self.short_title = self.short_title[i+1:]
             self.app.status.update(title=self.title)
             categories = set()
             for change in change_list:
                 categories |= set(change.getCategories())
             self.categories = sorted(categories)
+            self.chooseColumns()
+            self.header.update(self.categories)
             i = 0
             if self.reverse:
                 change_list.reverse()
@@ -271,9 +369,7 @@ class ChangeListView(urwid.WidgetWrap):
                     row = ChangeRow(self.app, change,
                                     prefixes.get(change.key),
                                     self.categories,
-                                    self.display_project,
-                                    self.display_owner,
-                                    self.display_updated,
+                                    self.enabled_columns,
                                     callback=self.onSelect)
                     self.listbox.body.insert(i, row)
                     self.change_rows[change.key] = row
@@ -288,12 +384,31 @@ class ChangeListView(urwid.WidgetWrap):
             else:
                 pos = min(focus_pos, len(self.listbox.body)-1)
             self.listbox.body.set_focus(pos)
-            if change_list:
-                self.header.update(self.categories)
         for key in unseen_keys:
             row = self.change_rows[key]
             del self.change_rows[key]
 
+    def chooseColumns(self):
+        currently_enabled_columns = self.enabled_columns.copy()
+        size = self.app.loop.screen.get_cols_rows()
+        cols = size[0]
+        for colinfo in COLUMNS:
+            if (colinfo.name not in self.disabled_columns):
+                cols -= colinfo.spacing
+        cols -= 3 * len(self.categories)
+
+        for colinfo in COLUMNS:
+            if colinfo.name in self.optional_columns:
+                if cols >= colinfo.spacing:
+                    self.enabled_columns.add(colinfo.name)
+                    cols -= colinfo.spacing
+                else:
+                    self.enabled_columns.discard(colinfo.name)
+        if currently_enabled_columns != self.enabled_columns:
+            self.header.updateColumns()
+            for key, value in six.iteritems(self.change_rows):
+                value.updateColumns()
+
     def getQueryString(self):
         if self.project_key is not None:
             return "project:%s %s" % (self.query_desc, self.app.config.project_change_list_query)
@@ -360,7 +475,7 @@ class ChangeListView(urwid.WidgetWrap):
         return (ret, prefixes)
 
     def clearChangeList(self):
-        for key, value in self.change_rows.iteritems():
+        for key, value in six.iteritems(self.change_rows):
             self.listbox.body.remove(value)
         self.change_rows = {}
 
@@ -390,6 +505,7 @@ class ChangeListView(urwid.WidgetWrap):
         with self.app.db.getSession() as session:
             change = session.getChange(change_key)
             change.reviewed = not change.reviewed
+            self.app.project_cache.clear(change.project)
             ret = change.reviewed
             reviewed_str = 'reviewed' if change.reviewed else 'unreviewed'
             self.log.debug("Set change %s to %s", change_key, reviewed_str)
@@ -424,17 +540,31 @@ class ChangeListView(urwid.WidgetWrap):
             self.listbox.focus_position = pos
 
     def keypress(self, size, key):
+        if self.searchKeypress(size, key):
+            return None
+
         if not self.app.input_buffer:
             key = super(ChangeListView, self).keypress(size, key)
         keys = self.app.input_buffer + [key]
         commands = self.app.config.keymap.getCommands(keys)
+        ret = self.handleCommands(commands)
+        if ret is True:
+            if keymap.FURTHER_INPUT not in commands:
+                self.app.clearInputBuffer()
+            return None
+        return key
+
+    def onResize(self):
+        self.chooseColumns()
+
+    def handleCommands(self, commands):
         if keymap.TOGGLE_LIST_REVIEWED in commands:
             self.unreviewed = not self.unreviewed
             self.refresh()
-            return None
+            return True
         if keymap.TOGGLE_REVIEWED in commands:
             if not len(self.listbox.body):
-                return None
+                return True
             pos = self.listbox.focus_position
             change_key = self.listbox.body[pos].change_key
             reviewed = self.toggleReviewed(change_key)
@@ -451,10 +581,10 @@ class ChangeListView(urwid.WidgetWrap):
                 # changes.
                 self.refresh()
                 self.advance()
-            return None
+            return True
         if keymap.TOGGLE_HIDDEN in commands:
             if not len(self.listbox.body):
-                return None
+                return True
             pos = self.listbox.focus_position
             change_key = self.listbox.body[pos].change_key
             hidden = self.toggleHidden(change_key)
@@ -469,10 +599,10 @@ class ChangeListView(urwid.WidgetWrap):
                 # where we're not just popping a row from the list of changes.
                 self.refresh()
                 self.advance()
-            return None
+            return True
         if keymap.TOGGLE_HELD in commands:
             if not len(self.listbox.body):
-                return None
+                return True
             pos = self.listbox.focus_position
             change_key = self.listbox.body[pos].change_key
             self.toggleHeld(change_key)
@@ -481,10 +611,10 @@ class ChangeListView(urwid.WidgetWrap):
                 change = session.getChange(change_key)
                 row.update(change, self.categories)
             self.advance()
-            return None
+            return True
         if keymap.TOGGLE_STARRED in commands:
             if not len(self.listbox.body):
-                return None
+                return True
             pos = self.listbox.focus_position
             change_key = self.listbox.body[pos].change_key
             self.toggleStarred(change_key)
@@ -493,10 +623,10 @@ class ChangeListView(urwid.WidgetWrap):
                 change = session.getChange(change_key)
                 row.update(change, self.categories)
             self.advance()
-            return None
+            return True
         if keymap.TOGGLE_MARK in commands:
             if not len(self.listbox.body):
-                return None
+                return True
             pos = self.listbox.focus_position
             change_key = self.listbox.body[pos].change_key
             row = self.change_rows[change_key]
@@ -505,10 +635,10 @@ class ChangeListView(urwid.WidgetWrap):
                 change = session.getChange(change_key)
                 row.update(change, self.categories)
             self.advance()
-            return None
+            return True
         if keymap.EDIT_TOPIC in commands:
             self.editTopic()
-            return None
+            return True
         if keymap.REFRESH in commands:
             if self.project_key:
                 self.app.sync.submitTask(
@@ -517,64 +647,73 @@ class ChangeListView(urwid.WidgetWrap):
                 self.app.sync.submitTask(
                     sync.SyncSubscribedProjectsTask(sync.HIGH_PRIORITY))
             self.app.status.update()
-            return None
+            return True
         if keymap.REVIEW in commands:
             rows = [row for row in self.change_rows.values() if row.mark]
             if not rows:
                 pos = self.listbox.focus_position
                 rows = [self.listbox.body[pos]]
             self.openReview(rows)
-            return None
+            return True
         if keymap.SORT_BY_NUMBER in commands:
             if not len(self.listbox.body):
-                return None
+                return True
             self.sort_by = 'number'
             self.clearChangeList()
             self.refresh()
-            return None
+            return True
         if keymap.SORT_BY_UPDATED in commands:
             if not len(self.listbox.body):
-                return None
+                return True
             self.sort_by = 'updated'
             self.clearChangeList()
             self.refresh()
-            return None
+            return True
         if keymap.SORT_BY_REVERSE in commands:
             if not len(self.listbox.body):
-                return None
+                return True
             if self.reverse:
                 self.reverse = False
             else:
                 self.reverse = True
             self.clearChangeList()
             self.refresh()
-            return None
+            return True
         if keymap.LOCAL_CHECKOUT in commands:
             if not len(self.listbox.body):
-                return None
+                return True
             pos = self.listbox.focus_position
             row = self.listbox.body[pos]
             self.app.localCheckoutCommit(row.project_name, row.commit_sha)
-            return None
+            return True
         if keymap.LOCAL_CHERRY_PICK in commands:
             if not len(self.listbox.body):
-                return None
+                return True
             pos = self.listbox.focus_position
             row = self.listbox.body[pos]
             self.app.localCherryPickCommit(row.project_name, row.commit_sha)
-            return None
+            return True
         if keymap.REFINE_CHANGE_SEARCH in commands:
             default = self.getQueryString()
             self.app.searchDialog(default)
-            return None
-        return key
+            return True
+        if keymap.ABANDON_CHANGE in commands:
+            self.abandonChange()
+            return True
+        if keymap.RESTORE_CHANGE in commands:
+            self.restoreChange()
+            return True
+        if keymap.INTERACTIVE_SEARCH in commands:
+            self.searchStart()
+            return True
+        return False
 
     def onSelect(self, button, change_key):
         try:
             view = view_change.ChangeView(self.app, change_key)
             self.app.changeScreen(view)
         except gertty.view.DisplayError as e:
-            self.app.error(e.message)
+            self.app.error(str(e))
 
     def openReview(self, rows):
         dialog = view_change.ReviewDialog(self.app, rows[0].current_revision_key)
@@ -624,3 +763,36 @@ class ChangeListView(urwid.WidgetWrap):
                         sync.SetTopicTask(change_key, sync.HIGH_PRIORITY))
         self.app.backScreen()
         self.refresh()
+
+    def abandonChange(self):
+        dialog = mywid.TextEditDialog(u'Abandon Change', u'Abandon message:',
+                                      u'Abandon Change', u'')
+        urwid.connect_signal(dialog, 'cancel', self.app.backScreen)
+        urwid.connect_signal(dialog, 'save', lambda button:
+                                 self.doAbandonRestoreChange(dialog, 'ABANDONED'))
+        self.app.popup(dialog)
+
+    def restoreChange(self):
+        dialog = mywid.TextEditDialog(u'Restore Change', u'Restore message:',
+                                      u'Restore Change', u'')
+        urwid.connect_signal(dialog, 'cancel', self.app.backScreen)
+        urwid.connect_signal(dialog, 'save', lambda button:
+                             self.doAbandonRestoreChange(dialog, 'NEW'))
+        self.app.popup(dialog)
+
+    def doAbandonRestoreChange(self, dialog, state):
+        rows = [row for row in self.change_rows.values() if row.mark]
+        if not rows:
+            pos = self.listbox.focus_position
+            rows = [self.listbox.body[pos]]
+        change_keys = [row.change_key for row in rows]
+        with self.app.db.getSession() as session:
+            for change_key in change_keys:
+                change = session.getChange(change_key)
+                change.status = state
+                change.pending_status = True
+                change.pending_status_message = dialog.entry.edit_text
+                self.app.sync.submitTask(
+                    sync.ChangeStatusTask(change_key, sync.HIGH_PRIORITY))
+        self.app.backScreen()
+        self.refresh()
